/* -*- mode: c; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4; c-file-style: "stroustrup"; -*-
 *
 * 
 *                This source code is part of
 * 
 *                 G   R   O   M   A   C   S
 * 
 *          GROningen MAchine for Chemical Simulations
 * 
 *                        VERSION 3.2.0
 * Written by David van der Spoel, Erik Lindahl, Berk Hess, and others.
 * Copyright (c) 1991-2000, University of Groningen, The Netherlands.
 * Copyright (c) 2001-2004, The GROMACS development team,
 * check out http://www.gromacs.org for more information.

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * If you want to redistribute modifications, please consider that
 * scientific software is very special. Version control is crucial -
 * bugs must be traceable. We will be happy to consider code for
 * inclusion in the official distribution, but derived work must not
 * be called official GROMACS. Details are found in the README & COPYING
 * files - if they are missing, get the official version at www.gromacs.org.
 * 
 * To help us fund GROMACS development, we humbly ask that you cite
 * the papers on the package - you can find them in the top README file.
 * 
 * For more info, check our website at http://www.gromacs.org
 * 
 * And Hey:
 * GROningen Mixture of Alchemy and Childrens' Stories
 */
/* This file is completely threadsafe - keep it that way! */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <math.h>
#include "typedefs.h"
#include "sysstuff.h"
#include "smalloc.h"
#include "macros.h"
#include "vec.h"
#include "futil.h"
#include "xvgr.h"
#include "gmx_fatal.h"
#include "bondf.h"
#include "copyrite.h"
#include "disre.h"
#include "main.h"
#include "mtop_util.h"

#include "do_fit.h" /* ZG */
#include "mdrun.h" /* ZG */


void init_disres(FILE *fplog,const gmx_mtop_t *mtop,
                 t_inputrec *ir,const t_commrec *cr,gmx_bool bPartDecomp,
                 t_fcdata *fcd,t_state *state)
{

    int          fa,nmol,i,npair,np;
    t_iparams    *ip;
    t_disresdata *dd;
    history_t    *hist;
    gmx_mtop_ilistloop_t iloop;
    t_ilist      *il;
    char         *ptr;

    dd = &(fcd->disres);

    if (gmx_mtop_ftype_count(mtop,F_DISRES) == 0)
    {
        dd->nres = 0;

        return;
    }
    
    if (fplog)
    {
        fprintf(fplog,"Initializing the distance restraints\n");fflush(fplog);
    }
    

    if (ir->eDisre == edrEnsemble)
    {
        gmx_fatal(FARGS,"Sorry, distance restraints with ensemble averaging over multiple molecules in one system are not functional in this version of GROMACS");
    }

    if (ir->eDisre == edrPairwise){
        dd->pairwise=TRUE;
    }

    dd->dr_weighting = ir->eDisreWeighting;
    dd->dr_fc        = ir->dr_fc;

    dd->popsize=ir->dr_popsize;
    
    if (EI_DYNAMICS(ir->eI))
    {
        dd->dr_tau   = ir->dr_tau;
    }
    else
    {
        dd->dr_tau   = 0.0;
    }
    if (dd->dr_tau == 0.0)
    {
        dd->dr_bMixed = FALSE;
        dd->ETerm = 0.0;
    }
    else
    {
        dd->dr_bMixed = ir->bDisreMixed;
        dd->ETerm = exp(-(ir->delta_t/ir->dr_tau));
    }
    dd->ETerm1        = 1.0 - dd->ETerm;
    
    ip = mtop->ffparams.iparams;

    dd->nres  = 0;
    dd->npair = 0;

    iloop = gmx_mtop_ilistloop_init(mtop);
    while (gmx_mtop_ilistloop_next(iloop,&il,&nmol)) {
        np = 0;
        for(fa=0; fa<il[F_DISRES].nr; fa+=3)
        {
            np++;
            npair = mtop->ffparams.iparams[il[F_DISRES].iatoms[fa]].disres.npair;
            if (np == npair)
            {
                dd->nres  += (ir->eDisre==edrEnsemble ? 1 : nmol)*npair;
                dd->npair += nmol*npair;
                np = 0;
            }
        }
    }

    if (cr && PAR(cr) && !bPartDecomp)
    {
        /* Temporary check, will be removed when disre is implemented with DD */
        const char *notestr="NOTE: atoms involved in distance restraints should be within the longest cut-off distance, if this is not the case mdrun generates a fatal error, in that case use particle decomposition (mdrun option -pd)";
        
        if (MASTER(cr))
            fprintf(stderr,"\n%s\n\n",notestr);
        if (fplog)
            fprintf(fplog,"%s\n",notestr);

        if (dd->dr_tau != 0 || ir->eDisre == edrEnsemble || cr->ms != NULL ||
            dd->nres != dd->npair)
        {
            gmx_fatal(FARGS,"Time or ensemble averaged or multiple pair distance restraints do not work (yet) with domain decomposition, use particle decomposition (mdrun option -pd)");
        }
        if (ir->nstdisreout != 0)
        {
            if (fplog)
            {
                fprintf(fplog,"\nWARNING: Can not write distance restraint data to energy file with domain decomposition\n\n");
            }
            if (MASTER(cr))
            {
                fprintf(stderr,"\nWARNING: Can not write distance restraint data to energy file with domain decomposition\n");
            }
            ir->nstdisreout = 0;
        }
    }

    snew(dd->rt,dd->npair);
    
    if (dd->dr_tau != 0.0)
    {
        hist = &state->hist;
        /* Set the "history lack" factor to 1 */
        state->flags |= (1<<estDISRE_INITF);
        hist->disre_initf = 1.0;
        /* Allocate space for the r^-3 time averages */
        state->flags |= (1<<estDISRE_RM3TAV);
        hist->ndisrepairs = dd->npair;
        snew(hist->disre_rm3tav,hist->ndisrepairs);
    }
    /* Allocate space for a copy of rm3tav,
     * so we can call do_force without modifying the state.
     */
    snew(dd->rm3tav,dd->npair);

    /* Allocate Rt_6 and Rtav_6 consecutively in memory so they can be
     * averaged over the processors in one call (in calc_disre_R_6)
     */
    snew(dd->Rt_6,2*dd->nres);
    dd->Rtav_6 = &(dd->Rt_6[dd->nres]);

    ptr = getenv("GMX_DISRE_ENSEMBLE_SIZE");
    if (cr && cr->ms != NULL && ptr != NULL)
    {
#ifdef GMX_MPI
        dd->nsystems = 0;
        sscanf(ptr,"%d",&dd->nsystems);
        if (fplog)
        {
            fprintf(fplog,"Found GMX_DISRE_ENSEMBLE_SIZE set to %d systems per ensemble\n",dd->nsystems);
        }
        check_multi_int(fplog,cr->ms,dd->nsystems,
                        "the number of systems per ensemble");
        if (dd->nsystems <= 0 ||  cr->ms->nsim % dd->nsystems != 0)
        {
            gmx_fatal(FARGS,"The number of systems %d is not divisible by the number of systems per ensemble %d\n",cr->ms->nsim,dd->nsystems);
        }
        /* Split the inter-master communicator into different ensembles */
        /* Keeping original method only if no pairwise averaging is requested - ZG */
        if (!dd->pairwise){
            MPI_Comm_split(cr->ms->mpi_comm_masters,
                           cr->ms->sim/dd->nsystems,
                           cr->ms->sim,
                           &dd->mpi_comm_ensemble);
        }
        if (fplog)
        {
            fprintf(fplog,"Our ensemble consists of systems:");
            for(i=0; i<dd->nsystems; i++)
            {
                fprintf(fplog," %d",
                        (cr->ms->sim/dd->nsystems)*dd->nsystems+i);
            }
            fprintf(fplog,"\n");
            }
        snew(dd->Rtl_6,dd->nres);
#endif
    }
    else
    {
        dd->nsystems = 1;
        dd->Rtl_6 = dd->Rt_6;
    }



    /* Intializing pairwise averaging stuff */
    if (dd->pairwise){
        snew(dd->leftmask,cr->ms->nsim);
        snew(dd->rightmask,cr->ms->nsim);
        /* Temporary variables to hold the values of Rt and Rtav summed with the left and right 
           neighbors of the node. Rt and Rtav will be summed in one call.
         */
        snew(dd->Rtleftmp,2*cr->ms->nsim*dd->nres); 
        snew(dd->Rtrightmp,2*cr->ms->nsim*dd->nres);
        int pn;
        int popsize=dd->popsize;
        int popstart=0;
        int popend=popsize-1;
        int leftn=0,rightn=0;

        for (pn=0; pn < cr->ms->nsim; pn++){
            // default mask: 0
            dd->leftmask[pn]=0;
            dd->rightmask[pn]=0;
        }

        for (pn=0; pn < cr->ms->nsim; pn++){
           if (pn > popend){popstart=popend;popend+=popsize;}
            if (cr->ms->sim == pn){
                dd->leftmask[pn]=1;
                dd->rightmask[pn]=1;
                leftn = (pn == popstart) ? popend : pn-1;
                dd->leftmask[leftn]=1;
                rightn = (pn == popend) ? popstart : pn+1;
                dd->rightmask[rightn]=1;
                //fprintf(stderr,"Node %d, leftn %d, rightn %d\n",cr->ms->sim,leftn,rightn);fflush(stderr);
            }
        }
        //fprintf(stderr,"Node %d, leftmask  %d%d%d%d, rightmask %d%d%d%d\n",cr->ms->sim,dd->leftmask[0],dd->leftmask[1],dd->leftmask[2],dd->leftmask[3],dd->rightmask[0],dd->rightmask[1],dd->rightmask[2],dd->rightmask[3]);fflush(stderr);
    }

    /* offical code below */
    if (dd->npair > 0)
    {
        if (fplog) {
            fprintf(fplog,"There are %d distance restraints involving %d atom pairs\n",dd->nres,dd->npair);
        }
        if (cr && cr->ms)
        {
            check_multi_int(fplog,cr->ms,fcd->disres.nres,
                            "the number of distance restraints");
        }
        please_cite(fplog,"Tropp80a");
        please_cite(fplog,"Torda89a");
        if (dd->pairwise){
            fprintf(fplog,"Performing distance restraint averaging for neighboring replicas using population size %d\n",cr->ms->nsim/dd->nsystems);
            please_cite(fplog,"Richter07"); // ZG
        }
    }
}

void calc_disres_R_6(const gmx_multisim_t *ms,
                     int nfa,const t_iatom forceatoms[],const t_iparams ip[],
                     const rvec x[],const t_pbc *pbc,
                     t_fcdata *fcd,history_t *hist)
{
    atom_id     ai,aj;
    int         fa,res,i,pair,ki,kj,m;
    int         type,npair,np;
    rvec        dx;
    real        *rt,*rm3tav,*Rtl_6,*Rt_6,*Rtav_6;
    real        rt_1,rt_3,rt2;
    ivec        it,jt,dt;
    t_disresdata *dd;
    real        ETerm,ETerm1,cf1=0,cf2=0,invn=0;
    gmx_bool        bTav;

    dd = &(fcd->disres);

    /* Variables for pairwise averaging - ZG */
    int rpn,popn;     
    real  *Rtleftmp, *Rtrightmp;
    if (dd->pairwise){
        Rtleftmp     = dd->Rtleftmp;
        Rtrightmp    = dd->Rtrightmp;
    }
 
    bTav         = (dd->dr_tau != 0);
    ETerm        = dd->ETerm;
    ETerm1       = dd->ETerm1;
    rt           = dd->rt;
    rm3tav       = dd->rm3tav;
    Rtl_6        = dd->Rtl_6;
    Rt_6         = dd->Rt_6;
    Rtav_6       = dd->Rtav_6;
    
    if (bTav)
    {
        /* scaling factor to smoothly turn on the restraint forces *
         * when using time averaging                               */
        dd->exp_min_t_tau = hist->disre_initf*ETerm;
        
        cf1 = dd->exp_min_t_tau;
        cf2 = 1.0/(1.0 - dd->exp_min_t_tau);
    }
    
    if (dd->nsystems > 1)
    {
        invn = 1.0/dd->nsystems;
    }

    /* 'loop' over all atom pairs (pair_nr=fa/3) involved in restraints, *
     * the total number of atoms pairs is nfa/3                          */
    res = 0;
    fa  = 0;
    while (fa < nfa)
    {
        type  = forceatoms[fa];
        npair = ip[type].disres.npair;
        
        Rtav_6[res] = 0.0;
        Rt_6[res]   = 0.0;
        
        /* Loop over the atom pairs of 'this' restraint */
        np = 0;
        while (fa < nfa && np < npair)
        {
            pair = fa/3;
            ai   = forceatoms[fa+1];
            aj   = forceatoms[fa+2];
            
            if (pbc)
            {
                pbc_dx_aiuc(pbc,x[ai],x[aj],dx);
            }
            else
            {
                rvec_sub(x[ai],x[aj],dx);
            }
            rt2  = iprod(dx,dx);
            rt_1 = gmx_invsqrt(rt2);
            rt_3 = rt_1*rt_1*rt_1;
            
            rt[pair]         = sqrt(rt2);
            if (bTav)
            {
                /* Here we update rm3tav in t_fcdata using the data
                 * in history_t.
                 * Thus the results stay correct when this routine
                 * is called multiple times.
                 */
                rm3tav[pair] = cf2*((ETerm - cf1)*hist->disre_rm3tav[pair] +
                                    ETerm1*rt_3);
            }
            else
            {
                rm3tav[pair] = rt_3;
            }

            Rt_6[res]       += rt_3*rt_3;
            Rtav_6[res]     += rm3tav[pair]*rm3tav[pair];

            /* MUMO TEST
            if ((res > 0) & (res <= 2)){
                Rt_6[res]=100*res+(ms->sim*ms->sim);
                Rtav_6[res]=1000*res+ms->sim;
            }
            MUMO TEST */

            fa += 3;
            np++;
        }
        if (dd->nsystems > 1){
            Rtl_6[res]   = Rt_6[res];
            if (dd->pairwise){
                for (rpn=0; rpn < ms->nsim; rpn++){
                    popn=(2*res*ms->nsim)+2*rpn;
                    /* Rtleftmp gets the value of this node at the position 
                       of this node and its left neighbor */
                    //fprintf(stderr,"node: %d | res: %d | rpn: %d | popn: %d | leftm: %d \n",ms->sim,res,rpn,popn,dd->leftmask[rpn]);fflush(stderr);
                    Rtleftmp[popn]=Rt_6[res]*dd->leftmask[rpn];
                    Rtrightmp[popn]=Rt_6[res]*dd->rightmask[rpn];
                    Rtleftmp[popn+1]=Rtav_6[res]*dd->leftmask[rpn];
                    Rtrightmp[popn+1]=Rtav_6[res]*dd->rightmask[rpn];
                }
            }
            /* Original code retained */
            else{
                Rt_6[res]   *= invn;
                Rtav_6[res] *= invn;
            }
        }

        res++;
    }


#ifdef GMX_MPI
    if (dd->nsystems > 1)
    {
        /* doing pairwise averaging by custom method, otherwise using original code*/
         if (dd->pairwise){
            gmx_sum_sim(2*ms->nsim*dd->nres,Rtleftmp,ms);
            gmx_sum_sim(2*ms->nsim*dd->nres,Rtrightmp,ms);
            for (res=0; res < dd->nres; res++){
                popn=(2*res*ms->nsim)+2*ms->sim;

               /* MUMO TEST  
               if ((res > 0) && (res < 2)){
                    fprintf(stderr,"PW Node %d, res %d, Rt_6   %f\n",ms->sim,res,Rt_6[res]); 
                    fprintf(stderr,"PW Node %d, res %d, Rtav_6 %f\n",ms->sim,res,Rtav_6[res]); 
                    fprintf(stderr,"PW Node %d, res %d, popn %d Leftmp %f Rightmp %f\n",ms->sim,res, popn,Rtleftmp[popn],Rtrightmp[popn]);fflush(stderr);

                }
                MUMO TEST */ 

                /* Rtleftmp now has the average value of this node and its left neighbor */
                Rt_6[res]=(Rtleftmp[popn]+Rtrightmp[popn])/4;
                Rtav_6[res]=(Rtleftmp[popn+1]+Rtrightmp[popn+1])/4;
            }
        }
        else{
            gmx_sum_comm(2*dd->nres,Rt_6,dd->mpi_comm_ensemble);
        }
    }
#endif

     /* MUMO TEST 
     for (res=1; res <= 2; res++){
         fprintf(stderr,"Node %d, res %d, Rt_6   %f\n",ms->sim,res,Rt_6[res]); 
         fprintf(stderr,"Node %d, res %d, Rtav_6 %f\n",ms->sim,res,Rtav_6[res]); 
    
     }
     MUMO TEST */
}

real ta_disres(int nfa,const t_iatom forceatoms[],const t_iparams ip[],
               const rvec x[],rvec f[],rvec fshift[],
               const t_pbc *pbc,const t_graph *g,
               real lambda,real *dvdlambda,
               const t_mdatoms *md,t_fcdata *fcd,
               int *global_atom_index)
{
    const real sixth=1.0/6.0;
    const real seven_three=7.0/3.0;
    
    atom_id     ai,aj;
    int         fa,res,npair,p,pair,ki=CENTRAL,m;
    int         type;
    rvec        dx;
    real        weight_rt_1;
    real        smooth_fc,Rt,Rtav,rt2,*Rtl_6,*Rt_6,*Rtav_6;
    real        k0,f_scal=0,fmax_scal,fk_scal,fij;
    real        tav_viol,instant_viol,mixed_viol,violtot,vtot;
    real        tav_viol_Rtav7,instant_viol_Rtav7;
    real        up1,up2,low;
    gmx_bool        bConservative,bMixed,bViolation;
    ivec        it,jt,dt;
    t_disresdata *dd;
    int         dr_weighting;
    gmx_bool        dr_bMixed;
    
    dd = &(fcd->disres);
    dr_weighting = dd->dr_weighting;
    dr_bMixed    = dd->dr_bMixed;
    Rtl_6        = dd->Rtl_6;
    Rt_6         = dd->Rt_6;
    Rtav_6       = dd->Rtav_6;

    tav_viol=instant_viol=mixed_viol=tav_viol_Rtav7=instant_viol_Rtav7=0;

    smooth_fc = dd->dr_fc;
    if (dd->dr_tau != 0)
    {
        /* scaling factor to smoothly turn on the restraint forces *
         * when using time averaging                               */
        smooth_fc *= (1.0 - dd->exp_min_t_tau); 
    }
    
    violtot = 0;
    vtot    = 0;
    
    /* 'loop' over all atom pairs (pair_nr=fa/3) involved in restraints, *
     * the total number of atoms pairs is nfa/3                          */
    res  = 0;
    fa   = 0;
    while (fa < nfa)
    {
        type  = forceatoms[fa];
        /* Take action depending on restraint, calculate scalar force */
        npair = ip[type].disres.npair;
        up1   = ip[type].disres.up1;
        up2   = ip[type].disres.up2;
        low   = ip[type].disres.low;
        k0    = smooth_fc*ip[type].disres.kfac;
        
        /* save some flops when there is only one pair */
        if (ip[type].disres.type != 2)
        {
            bConservative = (dr_weighting == edrwConservative) && (npair > 1);
            bMixed        = dr_bMixed;
            Rt   = pow(Rt_6[res],-sixth);
            Rtav = pow(Rtav_6[res],-sixth);
        }
        else
        {
            /* When rtype=2 use instantaneous not ensemble avereged distance */
            bConservative = (npair > 1);
            bMixed        = FALSE;
            Rt   = pow(Rtl_6[res],-sixth);
            Rtav = Rt;
        }
        
        if (Rtav > up1)
        {
            bViolation = TRUE;
            tav_viol = Rtav - up1;
        }
        else if (Rtav < low)
        {
            bViolation = TRUE;
            tav_viol = Rtav - low;
        }
        else
        {
            bViolation = FALSE;
        }
        
        if (bViolation)
        {
            /* NOTE:
             * there is no real potential when time averaging is applied
             */
            vtot += 0.5*k0*sqr(tav_viol);
            if (1/vtot == 0)
            {
                printf("vtot is inf: %f\n",vtot);
            }
            if (!bMixed)
            {
                f_scal   = -k0*tav_viol;
                violtot += fabs(tav_viol);
            }
            else
            {
                if (Rt > up1)
                {
                    if (tav_viol > 0)
                    {
                        instant_viol = Rt - up1;
                    }
                    else
                    {
                        bViolation = FALSE;
                    }
                }
                else if (Rt < low)
                {
                    if (tav_viol < 0)
                    {
                        instant_viol = Rt - low;
                    }
                    else
                    {
                        bViolation = FALSE;
                    }
                }
                else
                {
                    bViolation = FALSE;
                }
                if (bViolation)
                {
                    mixed_viol = sqrt(tav_viol*instant_viol);
                    f_scal     = -k0*mixed_viol;
                    violtot   += mixed_viol;
                }
            }
        }

        if (bViolation)
        {
            fmax_scal = -k0*(up2-up1);
            /* Correct the force for the number of restraints */
            if (bConservative)
            {
                f_scal  = max(f_scal,fmax_scal);
                if (!bMixed)
                {
                    f_scal *= Rtav/Rtav_6[res];
                }
                else
                {
                    f_scal /= 2*mixed_viol;
                    tav_viol_Rtav7     = tav_viol*Rtav/Rtav_6[res];
                    instant_viol_Rtav7 = instant_viol*Rt/Rt_6[res];
                }
            }
            else
            {
                f_scal /= (real)npair;
                f_scal  = max(f_scal,fmax_scal);
            }    
            
            /* Exert the force ... */
            
            /* Loop over the atom pairs of 'this' restraint */
            for(p=0; p<npair; p++)
            {
                pair = fa/3;
                ai   = forceatoms[fa+1];
                aj   = forceatoms[fa+2];
                
                if (pbc)
                {
                    ki = pbc_dx_aiuc(pbc,x[ai],x[aj],dx);
                }
                else
                {
                    rvec_sub(x[ai],x[aj],dx);
                }
                rt2 = iprod(dx,dx);
                
                weight_rt_1 = gmx_invsqrt(rt2);
                
                if (bConservative)
                {
                    if (!dr_bMixed)
                    {
                        weight_rt_1 *= pow(dd->rm3tav[pair],seven_three);
                    }
                    else
                    {
                        weight_rt_1 *= tav_viol_Rtav7*pow(dd->rm3tav[pair],seven_three)+
                            instant_viol_Rtav7*pow(dd->rt[pair],-7);
                    }
                }
                
                fk_scal  = f_scal*weight_rt_1;
                
                if (g)
                {
                    ivec_sub(SHIFT_IVEC(g,ai),SHIFT_IVEC(g,aj),dt);
                    ki=IVEC2IS(dt);
                }
                
                for(m=0; m<DIM; m++)
                {
                    fij            = fk_scal*dx[m];
                    
                    f[ai][m]      += fij;
                    f[aj][m]      -= fij;
                    fshift[ki][m] += fij;
                    fshift[CENTRAL][m] -= fij;
                }
                fa += 3;
            }
        }
        else
        {
            /* No violation so force and potential contributions */
            fa += 3*npair;
        }
        res++;
    }
    
    dd->sumviol = violtot;
    
    /* Return energy */
    return vtot;
}

void update_disres_history(t_fcdata *fcd,history_t *hist)
{
    t_disresdata *dd;
    int pair;
    
    dd = &(fcd->disres);
    if (dd->dr_tau != 0)
    {
        /* Copy the new time averages that have been calculated
         * in calc_disres_R_6.
         */
        hist->disre_initf = dd->exp_min_t_tau;
        for(pair=0; pair<dd->npair; pair++)
        {
            hist->disre_rm3tav[pair] = dd->rm3tav[pair];
        }
    }
}

/* Half-harmonic distance restraints with support for pairwise averaging over replicas - ZG */
void init_hhdisres(FILE *fplog,const gmx_mtop_t *mtop,
                 t_inputrec *ir,const t_commrec *cr,gmx_bool bPartDecomp,
                 const gmx_multisim_t *ms, // needed for pairwise averaging/population setup
                 t_fcdata *fcd,t_state *state)
{
    fprintf(stderr,"init_hhdisres\n");fflush(stderr);

    int          fa,nmol,i,npair,np;
    t_iparams    *ip;
    t_hhdisresdata *hhdd;
    history_t    *hist;
    gmx_mtop_ilistloop_t iloop;
    t_ilist      *il;
    char         *ptr;
    
    hhdd = &(fcd->hhdisres);

    if (gmx_mtop_ftype_count(mtop,F_HHDISRES) == 0)
    {
        hhdd->nres = 0;
        return;
    }
    
    if (fplog)
    {
        fprintf(fplog,"Initializing the half-harmonic distance restraints\n");
    }
    

    // Don't know what to to with this here - leaving as is for now
    if (ir->eHhdisre == ehhdrEnsemble)
    {
        gmx_fatal(FARGS,"Sorry, distance restraints with ensemble averaging over multiple molecules in one system are not functional in this version of GROMACS");
    }

    hhdd->hhdr_weighting = ir->eHhdisreWeighting;
    hhdd->hhdr_fc        = ir->hhdr_fc;
    hhdd->hhdr_exp       = ir->hhdr_exp;
    hhdd->pairwise=TRUE;

    if (EI_DYNAMICS(ir->eI))
    {
        hhdd->hhdr_tau   = ir->hhdr_tau;
    }
    else
    {
        hhdd->hhdr_tau   = 0.0;
    }
    if (hhdd->hhdr_tau == 0.0)
    {
        hhdd->hhdr_bMixed = FALSE;
        hhdd->ETerm = 0.0;
    }
    else
    {
        hhdd->hhdr_bMixed = ir->bHhdisreMixed;
        hhdd->ETerm = exp(-(ir->delta_t/ir->hhdr_tau)); // not sure we have hhdr_tau...
    }
    hhdd->ETerm1        = 1.0 - hhdd->ETerm;
    
    ip = mtop->ffparams.iparams;

    hhdd->nres  = 0;
    hhdd->npair = 0;
    iloop = gmx_mtop_ilistloop_init(mtop);
    while (gmx_mtop_ilistloop_next(iloop,&il,&nmol)) {
        np = 0;
        for(fa=0; fa<il[F_HHDISRES].nr; fa+=3)
        {
            np++;
            npair = mtop->ffparams.iparams[il[F_HHDISRES].iatoms[fa]].hhdisres.npair;
            if (np == npair)
            {
                hhdd->nres  += (ir->eHhdisre==ehhdrEnsemble ? 1 : nmol)*npair;
                hhdd->npair += nmol*npair;
                np = 0;
            }
        }
    }

    if (cr && PAR(cr) && !bPartDecomp)
    {
        /* Temporary check, will be removed when disre is implemented with DD */
        const char *notestr="NOTE: atoms involved in half-harmonic distance restraints should be within the longest cut-off distance, if this is not the case mdrun generates a fatal error, in that case use particle decomposition (mdrun option -pd)";
        
        if (MASTER(cr))
            fprintf(stderr,"\n%s\n\n",notestr);
        if (fplog)
            fprintf(fplog,"%s\n",notestr);

        if (hhdd->hhdr_tau != 0 || ir->eHhdisre == ehhdrEnsemble || cr->ms != NULL ||
            hhdd->nres != hhdd->npair)
        {
            gmx_fatal(FARGS,"Time or ensemble averaged or multiple pair distance restraints do not work (yet) with domain decomposition, use particle decomposition (mdrun option -pd)");
        }
        if (ir->nsthhdisreout != 0)
        {
            if (fplog)
            {
                fprintf(fplog,"\nWARNING: Can not write half-harmonic distance restraint data to energy file with domain decomposition\n\n");
            }
            if (MASTER(cr))
            {
                fprintf(stderr,"\nWARNING: Can not write half-harmonic distance restraint data to energy file with domain decomposition\n");
            }
            ir->nsthhdisreout = 0;
        }
    }

    snew(hhdd->rnij,hhdd->npair);
    snew(hhdd->rt,hhdd->npair);
    /* From here I deleted the history-related stuff... */    
    snew(hhdd->rm3tav,hhdd->npair);

    /* Allocate Rt_6 and Rtav_6 consecutively in memory so they can be
     * averaged over the processors in one call (in calc_disre_R_6)
     */
    snew(hhdd->Rtl_6,hhdd->nres);
    snew(hhdd->Rt_6,2*hhdd->nres);
    hhdd->Rtav_6 = &(hhdd->Rt_6[hhdd->nres]);

    hhdd->M      = ms->nsim;
    /* Here we set up the averaging over pairwise replicas */
    /* Variables for pairwise averaging, ZG */
    snew(hhdd->leftmask,hhdd->nres);
    snew(hhdd->rightmask,hhdd->nres);
    snew(hhdd->Rtleftmp,hhdd->M*hhdd->nres); // for summing left neighbors with leftmask
    snew(hhdd->Rtrightmp,hhdd->M*hhdd->nres); // for summing right neighbors with rightmask
    snew(hhdd->Rt_left,hhdd->nres);
    snew(hhdd->Rt_right,hhdd->nres);
    snew(hhdd->viol,hhdd->nres);
    snew(hhdd->violl,hhdd->nres);
    snew(hhdd->violr,hhdd->nres);



    fprintf(stderr,"init_hhdisres: npair: %d, nres: %d\n",hhdd->npair, hhdd->nres);fflush(stderr);

    //Doing HHDisre calculation for population size pop
    if(ir->hhdrpopsize != -1){
       hhdd->pop=ir->hhdrpopsize;
       //sd->popnum=sd->M % sd->pop;
       if (((hhdd->M % hhdd->pop) != 0) || (hhdd->pop <= 1)){
           fprintf(stderr,"WARNING: incompatible population size for half-harmonic distance restraints: pop %d, nodes %d - setting pop to nodes\n",hhdd->pop,hhdd->M);fflush(stderr);
           hhdd->pop=hhdd->M; // fallback
           hhdd->popnum=1;
       }
       else{
          hhdd->popnum=hhdd->M/hhdd->pop;
       }
    }
    else{
      hhdd->pop=hhdd->M; // if popsize not set, default all nodes
      hhdd->popnum=1;
    }

    fprintf(fplog,"Doing half-harmonic distance restraint calc for %d populations of size %d\n",hhdd->popnum,hhdd->pop);
    fprintf(stderr,"Doing half-harmonic distance restraint calc for %d populations of size %d\n",hhdd->popnum,hhdd->pop);fflush(stderr);

    /* setting local population */
     int pn;
     int apop=0;
     int pm=0;
     for (pn=0; pn < hhdd->M; pn++){
       if (ms->sim == pn){
         hhdd->popid=apop;
       }
       pm++;
       if (pm == hhdd->pop){apop++;pm=0;}
     }
 
    /* generating averaging mask for left and right neighbors */
    /* index for first and last members of population */
    int popstart=0;
    int popend=hhdd->pop-1;
    int leftn=0,rightn=0;

    //snew(hhdd->leftmask,hhdd->M);
    //snew(hhdd->rightmask,hhdd->M);

    for (pn=0; pn < hhdd->M; pn++){
        // default mask: 0
        if (hhdd->pairwise){
            hhdd->leftmask[pn]=0;
            hhdd->rightmask[pn]=0;
            if (pn > popend){popstart=popend;popend+=hhdd->pop;}
            if (ms->sim == pn){
                hhdd->leftmask[pn]=1;
                hhdd->rightmask[pn]=1;
                leftn = (pn == popstart) ? popend : pn-1;
                hhdd->leftmask[leftn]=1;
                hhdd->leftn=leftn;
                rightn = (pn == popend) ? popstart : pn+1;
                hhdd->rightmask[rightn]=1;
                hhdd->rightn=rightn;
            }
        }// if pairwise
        else{ // full ensemble averaging
            hhdd->leftmask[pn]=1;
            hhdd->rightmask[pn]=1;
            hhdd->leftn=0;
            hhdd->rightn=0;
        }
    }

    fprintf(fplog,"Replica %d: population id %d, left neighbor: %d, right neighbor: %d\n",ms->sim,hhdd->popid,hhdd->leftn,hhdd->rightn);

    /* GZ */

    if (hhdd->npair > 0)
    {
        if (fplog) {
            fprintf(fplog,"There are %d half-harmonic distance restraints involving %d atom pairs\n",hhdd->nres,hhdd->npair);
        }
        if (cr && cr->ms)
        {
            check_multi_int(fplog,cr->ms,fcd->hhdisres.nres,
                            "the number of distance restraints");
        }
        please_cite(fplog,"Tropp80a");
        please_cite(fplog,"Torda89a");
        please_cite(fplog,"Richter2007");
    }
}


void calc_hhdisres(const gmx_multisim_t *ms,
                     int nfa,const t_iatom forceatoms[],const t_iparams ip[],
                     const rvec x[],const t_pbc *pbc,
                     t_fcdata *fcd,history_t *hist)
{
    fprintf(stderr,"calc_hhdisres\n");fflush(stderr);

    atom_id     ai,aj;
    int         fa,res,i,pair,ki,kj,m;
    int         type,npair,np;
    rvec        dx;
    real        *rt,*rm3tav,*Rtl_6,*Rt_6,*Rtav_6;
    real        *Rt_left, *Rt_right, *Rtleftmp, *Rtrightmp, *RNij, rav_3;
    real        rt_1,rt_3,rt2;
    ivec        it,jt,dt;
    t_hhdisresdata *hhdd;
    real        ETerm,ETerm1,cf1=0,cf2=0,invn=0;
    gmx_bool        bTav;
    int          exp;
    real         rtleft, rtright;
    int          rpn,hhpopn;

    hhdd = &(fcd->hhdisres);
    bTav         = (hhdd->hhdr_tau != 0);
    ETerm        = hhdd->ETerm;
    ETerm1       = hhdd->ETerm1;
    rt           = hhdd->rt;
    rm3tav       = hhdd->rm3tav;
    Rt_6         = hhdd->Rt_6;
    Rt_left      = hhdd->Rt_left;
    Rt_right     = hhdd->Rt_right;
    Rtleftmp     = hhdd->Rtleftmp;
    Rtrightmp    = hhdd->Rtrightmp;
    Rtav_6       = hhdd->Rtav_6;
    exp          = hhdd->hhdr_exp;
    RNij         = hhdd->rnij; // here we set the value used in the force calc

    exp=6; // for now 

    /* 'loop' over all atom pairs (pair_nr=fa/3) involved in restraints, *
     * the total number of atoms pairs is nfa/3                          */
    res = 0;
    fa  = 0;
    while (fa < nfa)
    {

        type  = forceatoms[fa];
        npair = ip[type].hhdisres.npair;
        
        Rtav_6[res] = 0.0;
        Rt_6[res]   = 0.0;
        
        /* Loop over the atom pairs of 'this' restraint */
        np = 0;

       real rnij_x=0;
       real rnij_y=0;
       real rnij_z=0;



        while (fa < nfa && np < npair)
        {
            //fprintf(stderr,"fa: %d | res: %d (npair: %d)\n",fa,res,npair);fflush(stderr);
            pair = fa/3;
            ai   = forceatoms[fa+1];
            aj   = forceatoms[fa+2];
            
            fprintf(stderr,"ai: %d aj: %d\n",ai,aj);fflush(stderr);

            if (pbc)
            {
                pbc_dx_aiuc(pbc,x[ai],x[aj],dx);
            }
            else
            {
                rvec_sub(x[ai],x[aj],dx);
            }
            rt2  = iprod(dx,dx);
            rt_1 = gmx_invsqrt(rt2);
            rt_3 = rt_1*rt_1*rt_1;
            
            //rt[pair]         = sqrt(rt2);
            /* history stuff elminiated */
            //rm3tav[pair] = rt_3;
            /* Code purged to eliminate 3 names for the same value (rt_3) */
            //Rtl_6[res]       += rt_3*rt_3;
            //Rtav_6[res]     += rm3tav[pair]*rm3tav[pair];

            rnij_x+=dx[0]; rnij_y+=dx[1]; rnij_z+=dx[2];

            // There is some controversy about the exponent, let the user decide.
            // A revision of averaging might be necessary, however...


            if (exp == 3)
	        Rtav_6[res] += rt_3;
            else if (exp == 6)
	        Rtav_6[res] += rt_3*rt_3;

            fa += 3;
            np++;

        }
        RNij[res]=sqrt(rnij_x*rnij_x+rnij_y*rnij_y+rnij_z*rnij_z);

  
        for (rpn=0; rpn < hhdd->M; rpn++){
            hhpopn=(res*hhdd->M)+rpn;
            //fprintf(stderr,"exp: %d | res: %d | hpopn: %d, leftm: %d \n",exp,res,hhpopn,hhdd->leftmask[rpn]);fflush(stderr);
            Rtleftmp[hhpopn]=Rtav_6[res]*hhdd->leftmask[rpn];
            Rtrightmp[hhpopn]=Rtav_6[res]*hhdd->rightmask[rpn];
        }
        res++;
    
    }// while fa


#ifdef GMX_MPI
    if (hhdd->M > 1)
    {
        if (hhdd->pairwise){
            //gmx_sum_sim(sd->nres,x2,ms);
            gmx_sum_sim(hhdd->M*hhdd->nres,Rtleftmp,ms);
            gmx_sum_sim(hhdd->M*hhdd->nres,Rtrightmp,ms);
            for (res=0; res < hhdd->nres; res++){
                hhpopn=(res*hhdd->M)+ms->sim;
                Rt_left[res]=(Rtav_6[res]+(Rtleftmp[hhpopn]/2)/2);
                Rt_right[res]=(Rtav_6[res]+(Rtrightmp[hhpopn]/2)/2);
            }
        }// pairwise (population-based) averaging
        else{
            gmx_sum_sim(hhdd->nres,Rtav_6,ms);
        }//overall averaging
    }


#endif
  fprintf(stderr,"end of calc hhdisres\n");fflush(stderr);

}//calc_hhdisres

real ta_hhdisres(int nfa,const t_iatom forceatoms[],const t_iparams ip[],
               const rvec x[],rvec f[],rvec fshift[],
               const t_pbc *pbc,const t_graph *g,
               real lambda,real *dvdlambda,
               const t_mdatoms *md,t_fcdata *fcd,
               int *global_atom_index)
{

  const real sixth=1.0/6.0;
  const real seven_three=7.0/3.0;

  fprintf(stderr,"in hh_tadisres\n");fflush(stderr);
  
  atom_id     ai,aj;
  int         fa,res,npair,p,pair,ki=CENTRAL,m;
  int         type,label;
  rvec        dx;
  real        weight_rt_1;
  real        smooth_fc,Rt,Rav,rt2,*Rtl_6,*Rt_6,*Rtav_6, *RNij;
  real        k0,f_scal=0,fmax_scal,fk_scal,fij;
  real        tav_viol,instant_viol,mixed_viol,violtot,vtot;
  real        tav_viol_Rav7,instant_viol_Rav7;
  real        up1,up2,low;
  gmx_bool        bConservative,bMixed,bViolation;
  ivec        it,jt,dt;
  t_hhdisresdata *hhdd;
  real        dr_fc;
  real        *viol, *violl, *violr;

  hhdd = &(fcd->hhdisres);
  //dr_weighting = hhdd->hhdr_weighting;
  //dr_bMixed    = hhdd->dr_bMixed;
  dr_fc        = hhdd->hhdr_fc;
  Rtl_6        = hhdd->Rtl_6;
  Rt_6         = hhdd->Rt_6;
  Rtav_6       = hhdd->Rtav_6;
  viol         = hhdd->viol;
  violl        = hhdd->violl;
  violr        = hhdd->violr;
  RNij         = hhdd->rnij;

  real oneperexp=1.0/(real)hhdd->hhdr_exp;
  real *Rt_left,*Rt_right;
  Rt_left        = hhdd->Rt_left;
  Rt_right       = hhdd->Rt_right;
  real r;

  violtot = 0;
  vtot    = 0;

  hhdd->rho=0.0;

      res  = 0;
    fa   = 0;
    while (fa < nfa)
    {
        type  = forceatoms[fa];
        /* Take action depending on restraint, calculate scalar force */
        npair = ip[type].hhdisres.npair;
        up1   = ip[type].hhdisres.up1;
        up2   = ip[type].hhdisres.up2;
        low   = ip[type].hhdisres.low;
        //k0    = smooth_fc*ip[type].hhdisres.kfac;
        
        if (hhdd->pairwise){

            Rt_left[res] = pow(Rt_left[res],-oneperexp);
            Rt_right[res] = pow(Rt_right[res],-oneperexp);
            violl[res]=Rt_left[res] > up1 ? Rt_left[res]-up1 : Rt_left[res] < low ? low-Rt_left[res] : 0.0;
            hhdd->rho += 0.5*violl[res]*violl[res];

            //fprintf(stderr,"left (%d): %7.3f up1: %7.3f low: %7.3f  viol: %7.3f\n",res,Rt_left[res],up1,low,violl[res]);fflush(stdout);

            violr[res]=Rt_right[res] > up1 ? Rt_right[res]-up1 : Rt_right[res] < low ? low-Rt_right[res] : 0.0;
            hhdd->rho += 0.5*violr[res]*violr[res];
            //fprintf(stderr,"right (%d): %7.3f up1: %7.3f low: %7.3f  viol: %7.3f\n",res,Rt_right[res],up1,low,violr[res]);fflush(stdout);

        }
        else{
            Rtav_6[res] = pow(Rtav_6[res],-oneperexp);
            viol[res]=Rtav_6[res] > up1 ? Rtav_6[res]-up1 : Rtav_6[res] < low ? low-Rtav_6[res] : 0.0;
            //fprintf(stderr,"rav6 (%d): %7.3f up1: %7.3f low: %7.3f  viol: %7.3f\n",res,Rav_6[res],up1,low,viol[res]);fflush(stdout);

            hhdd->rho += (1.0/hhdd->M)*viol[res]*viol[res];
	
        }

        fa += 3*npair;
        res++;

    }// while fa < nfa
            
     //fprintf(stderr,"HHDR> Rho = %7.5f, Rhomin = %7.5f\n",dd->rho,dd->rhomin);fflush(stderr);

    vtot=0.0;
    // No rhomin for the 1st calculation, so skip the main part
    if (hhdd->rhomin < 0.0){
        hhdd->rhomin=hhdd->rho;
    } else if (hhdd->rho > hhdd->rhomin){
        vtot=(hhdd->hhdr_fc*hhdd->M/2.0)*(hhdd->rho-hhdd->rhomin)*(hhdd->rho-hhdd->rhomin);

    
    // Here comes the force

	fa=0;
	res=0;

    // variables as in hqbm.src
	float pcor,pcor1,pcor2;
	float aux,aux1,aux2;
	float bux=2.0*(hhdd->rho-hhdd->rhomin)*hhdd->hhdr_fc/(float)hhdd->nres;
	float avdis,rnij;
	float fap;
	float onepersix=-1.0/6.0;

	while (fa < nfa){

        type  = forceatoms[fa];
        npair = ip[type].hhdisres.npair;
        up1   = ip[type].hhdisres.up1;
        up2   = ip[type].hhdisres.up2;
        low   = ip[type].hhdisres.low;

        // Variable name conversion :-)
	    rnij=pow(Rtl_6[res],-sixth);
	    if (hhdd->pairwise){
		if (hhdd->hhdr_exp == 6){
		    pcor1=pow(Rt_left[res]/rnij,7.0);
		    pcor2=pow(Rt_right[res]/rnij,7.0);
		}
		else if (hhdd->hhdr_exp == 3){
		    pcor1=pow(Rt_left[res]/rnij,4.0);
		    pcor2=pow(Rt_right[res]/rnij,4.0);
		}
		aux1=bux*pcor1*(violl[res])*pow(rnij,7.0);
		aux2=bux*pcor2*(violr[res])*pow(rnij,7.0);
		fprintf(stderr,"rnij = %7.3f | pcor1 = %7.3f | pcor2 = %7.3f | aux1 = %7.3f | aux2 = %7.3f \n",rnij,pcor1,pcor2,aux1,aux2);fflush(stderr); 

	    }
	    else{
		if (hhdd->hhdr_exp == 6){
		    pcor=pow(Rtav_6[res]/rnij,7.0);
		}
		else if (hhdd->hhdr_exp == 3){
		    pcor=pow(Rtav_6[res]/rnij,4.0);
		}
		aux=bux*pcor*(viol[res])*pow(rnij,7.0);
	    }
            
        /* Loop over the atom pairs of 'this' restraint */
        for(p=0; p<npair; p++)
        {
            pair = fa/3;
            ai   = forceatoms[fa+1];
            aj   = forceatoms[fa+2];
                
            if (pbc)
            {
                ki = pbc_dx_aiuc(pbc,x[ai],x[aj],dx);
            }
            else
            {
                rvec_sub(x[ai],x[aj],dx);
            }
                
            if (g)
            { 
                ivec_sub(SHIFT_IVEC(g,ai),SHIFT_IVEC(g,aj),dt);
                ki=IVEC2IS(dt);
            }
                
            for(m=0; m<DIM; m++)
            {
                if (hhdd->pairwise){
                    fij            = aux1*dx[m]*pow(r,-8.0) + aux2*dx[m]*pow(r,-8.0);
                }
	            else{
                    fij            = aux*dx[m]*pow(r,-8.0);
                }
                fprintf(stderr,"RES: %d FORCE[%d]: %7.3f\n",res,m,fij);fflush(stderr);
                    
                f[ai][m]      += fij;
                f[aj][m]      -= fij;
                fshift[ki][m] += fij;
                fshift[CENTRAL][m] -= fij;
             }
            fa += 3;
        }

        res++;
    }
    
    hhdd->sumviol = violtot;
    
    }// if rho > rhomin
    /* Return energy */
    return vtot;

}// ta_hhdisres


/* S2 restraining stuff by ZG */
void init_s2res (FILE *fplog, const gmx_mtop_t *mtop, 
                 t_inputrec *ir, const t_commrec *cr, 
                 rvec xref[], const gmx_multisim_t *ms, // these two from init_orires
                 t_fcdata *fcd, t_state *state)
{
    //fprintf(stderr,"Init S2 res\n");

    // from disres:
    int          fa,nmol,i,npair,np;
    t_iparams    *ip;
    t_s2resdata  *sd;
    history_t    *hist;
    gmx_mtop_ilistloop_t iloop;
    t_ilist      *il;
    char         *ptr;
    // from orires.c:
    rvec         com;
    double       mtot;
    gmx_mtop_atomloop_all_t aloop;
    t_atom *atom;
    int    j,d,ex,nr,*nr_ex; // not sure all of these needed, will check

    sd = &(fcd->s2res);

    if (gmx_mtop_ftype_count(mtop,F_S2RES) == 0)
    {
        sd->nres = 0;
        return;
    }
    
    if (fplog)
    {
        fprintf(fplog,"Initializing the S2 order parameter restraints\n");
    }
    ip = mtop->ffparams.iparams;

    sd->s2step=-1;
    sd->nres  = 0;
    sd->npair = 0;
    iloop = gmx_mtop_ilistloop_init(mtop);
    while (gmx_mtop_ilistloop_next(iloop,&il,&nmol)) {
        np = 0;
        for(fa=0; fa<il[F_S2RES].nr; fa+=3)
        {
            np++;
            npair = mtop->ffparams.iparams[il[F_S2RES].iatoms[fa]].s2res.npair;
            if (np == npair)
            {
                sd->nres  += npair; // I do not know whether this ensemble stuff was functional here (think not)
                                    // but I will not bother with it now.
                sd->npair += nmol*npair;
                np = 0;
            }
        }
    }

    sd->alpha=abs(ir->s2res_fc);
    if (ir->s2res_fc < 0){sd->do_fit=0;}
    else {sd->do_fit=1;}

    fprintf(stderr,"FITTING: do_fit= %d \n",sd->do_fit);
    //sd->do_fit = (sd->alpha == ir->s2res_fc); // old version kept as reference
    //fprintf(stderr,"alpha= %7.3f  s2res_fc = %7.3f\n",sd->alpha, ir->s2res_fc);
    snew(sd->ijvector,sd->nres);
    snew(sd->s2calc,sd->nres);
    snew(sd->s2diff,sd->nres);
    snew(sd->s2exp,sd->nres);
    snew(sd->s2err,sd->nres);
    snew(sd->rijmin4,sd->nres);
    snew(sd->x2,sd->nres);
    snew(sd->y2,sd->nres);
    snew(sd->z2,sd->nres);
    snew(sd->xy,sd->nres);
    snew(sd->xz,sd->nres);
    snew(sd->yz,sd->nres);
    sd->rhomin=-1.0;
    sd->M=ms->nsim; 
    sd->s2reout=ir->nsts2reout;

    //fprintf(stderr,"pop size: %d\n",ir->s2popsize);fflush(stderr);

    //Doing S2 calculation for population size pop
    if(ir->s2popsize != -1){
       sd->pop=ir->s2popsize;
       //sd->popnum=sd->M % sd->pop;
       if (((sd->M % sd->pop) != 0) || (sd->pop == 1)){
           fprintf(stderr,"WARNING: incompatible popsulation size: pop %d, nodes %d - setting pop to nodes\n",sd->pop,sd->M);
           sd->pop=sd->M; // fallback
           sd->popnum=1;
       }
       else{
          sd->popnum=sd->M/sd->pop;
       }
    }
    else{
      sd->pop=sd->M; // if popsize not set, default all nodes
      sd->popnum=1;
    }

    fprintf(fplog,"Doing S2 calc for %d populations of size %d\n",sd->popnum,sd->pop);
    fprintf(stderr,"Doing S2 calc for %d populations of size %d\n",sd->popnum,sd->pop);

     snew(sd->popsum,sd->popnum);
     // setting local population
     int pn;
     int apop=0;
     int pm=0;
     for (pn=0; pn < sd->M; pn++){
       if (ms->sim == pn){
         sd->popid=apop;
       }
       pm++;
       if (pm == sd->pop){apop++;pm=0;}
     }
     // setting popsum 1 for local population, 0 for all others (=adding node specificity)
     for (pn=0; pn < sd->popnum; pn++){ // watch out, temp variable pn is not as above
       if (pn==sd->popid){sd->popsum[pn]=1;}
       else{sd->popsum[pn]=0;}
     }

     fprintf(fplog,"S2 population for node %d is %d\n",ms->sim,sd->popid);
     fprintf(stderr,"S2 population for node %d is %d\n",ms->sim,sd->popid);

     // creating an array to hold all components for all populations
     // this allows to keep the multireplica summing code below
     // values for non-self populations will be 0
     snew(sd->px2,sd->nres*sd->popnum);
     snew(sd->py2,sd->nres*sd->popnum);
     snew(sd->pz2,sd->nres*sd->popnum);
     snew(sd->pxy,sd->nres*sd->popnum);
     snew(sd->pxz,sd->nres*sd->popnum);
     snew(sd->pyz,sd->nres*sd->popnum);
     /* end of population stuff */


    /* Bothering with this all only if fit needed */ 
    /* fit code from orires.c (trying to follow the new version... */
    if (sd->do_fit == 1){
        sd->nref = 0;
        for(i=0; i<mtop->natoms; i++)
        {
            if (ggrpnr(&mtop->groups,egcS2RFIT,i) == 0)
            {
                sd->nref++;
            }
        }
        snew(sd->mref,sd->nref);
        snew(sd->xref,sd->nref);
        snew(sd->xtmp,sd->nref);

        /* Determine the reference structure on the master node.
         * Copy it to the other nodes after checking multi compatibility,
         * so we are sure the subsystems match before copying.
         */
        clear_rvec(com);
        mtot = 0.0;
        j = 0;
        aloop = gmx_mtop_atomloop_all_init(mtop);
        while(gmx_mtop_atomloop_all_next(aloop,&i,&atom))
        {
            if (mtop->groups.grpnr[egcS2RFIT] == NULL ||
                mtop->groups.grpnr[egcS2RFIT][i] == 0)
            {
                /* Not correct for free-energy with changing masses */
                sd->mref[j] = atom->m;
                if (ms==NULL || MASTERSIM(ms))
                {
                    copy_rvec(xref[i],sd->xref[j]);
                    for(d=0; d<DIM; d++)
                    {
                        com[d] += sd->mref[j]*xref[i][d];
                    }
                }
                mtot += sd->mref[j];
                j++;
            }
        }//while gmx_top

    }// if do_fit

    svmul(1.0/mtot,com,com);
    if (ms==NULL || MASTERSIM(ms))
    {
        for(j=0; j<sd->nref; j++)
        {
            rvec_dec(sd->xref[j],com);
        }
    }

    if (sd->npair > 0)
    {
        if (fplog) {
            fprintf(fplog,"There are %d S2 restraints involving %d atom pairs\n",sd->nres,sd->npair);
        }
        if (cr && cr->ms)
        {
            check_multi_int(fplog,cr->ms,fcd->s2res.nres,
                            "the number of S2 restraints");
        }
        if (sd->do_fit == 1){
            fprintf(fplog,"  the fit group for S2 restraints consists of %d atoms and has total mass %g\n",
                    sd->nref,mtot);
                    /* Copy the reference coordinates from the master to the other nodes */
                    gmx_sum_sim(DIM*sd->nref,sd->xref[0],ms);
        }

        please_cite(fplog,"Best04");
        please_cite(fplog,"Fizil15");
    }


}

void calc_s2res(FILE *fplog, 
                      const gmx_multisim_t *ms,
                     int nfa,const t_iatom forceatoms[],const t_iparams ip[],
                     const t_mdatoms *md,                                       // from orires
                //const t_commrec *cr,
                     const rvec x[],const t_pbc *pbc,
                     t_fcdata *fcd,history_t *hist
                //, int step
                      )
{
    //fprintf(stderr,"Calc S2 res\n");fflush(stderr);
    t_s2resdata *sd;
    sd=&(fcd->s2res);
    sd->nres=nfa/3;
    atom_id ai,aj;
    rvec dx, fx; /* Original and fitted interatomic vectors */
    rvec *ijv;
    ijv=sd->ijvector;
    //int testatom1=18,testatom2=25;
    int type, m, n, d;
    real *s2calc = sd->s2calc;
    real S2;
    real invn;
    double mtot;
    real rijmin=0, bondlen, actual;
    real *rijmin4=sd->rijmin4;

    int fa=0, pair=0, npair;
    int ndn;
    tensor R;

    real *x2=sd->x2;
    real *y2=sd->y2;
    real *z2=sd->z2;
    real *xy=sd->xy;
    real *xz=sd->xz;
    real *yz=sd->yz;

    // population stuff
    real *px2=sd->px2;
    real *py2=sd->py2;
    real *pz2=sd->pz2;
    real *pxy=sd->pxy;
    real *pxz=sd->pxz;
    real *pyz=sd->pyz;
    int pn;
    int ppair;
    real pinvn;
    int nrpop=sd->nres*sd->popnum;
    // population suff

    int nref = sd->nref;
    real *mref = sd->mref;
    rvec *xref = sd->xref;
    rvec *xtmp = sd->xtmp;
    rvec com;
    int i,j;


    // Reading in experimental values at first invocation to avoid re-read every time
    if (sd->rhomin < 0.0){
        fa  = 0;
        while (fa < nfa)
        {
            pair=fa/3;
            type  = forceatoms[fa];
            sd->s2exp[pair]    = ip[type].s2res.S2;
            sd->s2err[pair]    = ip[type].s2res.S2err;
            fa+=3;
        }
    }

    // My own counter until I figure out everything about energy logging...
    sd->s2step++;

    // Calculating rotation matrix and doing rotation if needed - code from orires.c
    if (sd->do_fit){
        clear_rvec(com);
        mtot = 0;
        j=0;
        for(i=0; i<md->nr; i++)
        {
            if (md->cS2RF[i] == 0)
            {
                copy_rvec(x[i],xtmp[j]);
                mref[j] = md->massT[i];
                for(d=0; d<DIM; d++)
                {
                    com[d] += mref[j]*xref[j][d];
                    //fprintf(stderr,"d= %d  j= %d xref[j][d]=%7.3f\n",j,d,xref[j][d]);fflush(stderr);
                }
                mtot += mref[j];
                j++;
            }
        }

        //fprintf(stderr,"R mtx coming down\n");fflush(stderr);
        svmul(1.0/mtot,com,com);
        for(j=0; j<nref; j++)
        {
            rvec_dec(xtmp[j],com);
        }
        /* Calculate the rotation matrix to rotate x to the reference orientation */
        calc_fit_R(DIM,nref,mref,xref,xtmp,R);
        copy_mat(R,sd->R);

    }// if do_fit

    if (sd->M > 1)
    {
        invn = 1.0/sd->M;
        pinvn=1.0/sd->pop;
    }


    fa  = 0;
    while (fa < nfa)
    {
        pair=fa/3;
        type  = forceatoms[fa];
        npair = ip[type].s2res.npair;
	    ai   = forceatoms[fa+1];
	    aj   = forceatoms[fa+2];
	    if (pbc)
            pbc_dx_aiuc(pbc,x[aj],x[ai],dx);
	    else
            rvec_sub(x[aj],x[ai],dx);

        bondlen    = ip[type].s2res.bondlen;
	    actual=norm(dx);

        //fprintf(stderr,"xaj %7.3f xai %7.3f invn %7.3f sdM %d \n",x[aj][0],x[ai][0],invn,sd->M);fflush(stderr);

	   if ((bondlen == -1.0) && (sd->rhomin == -1)){
		rijmin=norm(dx);
		rijmin4[pair]=rijmin*rijmin*rijmin*rijmin;
	   } else if (bondlen == 0.0){
		rijmin=norm(dx);
		rijmin4[pair]=rijmin*rijmin*rijmin*rijmin;
	   } else if (bondlen > 0.0){
	        rijmin=bondlen;
		rijmin4[pair]=rijmin*rijmin*rijmin*rijmin;
	   }

	   if (sd->do_fit){
	     mvmul(R,dx,fx);
	     copy_mat(R,sd->R); // Guess that this is not needed here as already done
	   } else {
           /* Go on with fx even if there is no fit -> so copy dx to it */
	     copy_rvec(dx,fx);  
	   }
	    copy_rvec(fx,ijv[pair]);  

	    x2[pair]=fx[0]*fx[0]*invn;
	    y2[pair]=fx[1]*fx[1]*invn;
	    z2[pair]=fx[2]*fx[2]*invn;
	    xy[pair]=fx[0]*fx[1]*invn;
	    xz[pair]=fx[0]*fx[2]*invn;
	    yz[pair]=fx[1]*fx[2]*invn;

        // population stuff goes here
        // creating a variable for each population, but will be 0 for
        // all nonlocal populations
        // 0..popnum-1: 1st pair, popnum..2*popnum-1: 2nd pair etc.
        ppair=pair*sd->popnum;
        for (pn=0; pn < sd->popnum; pn++){
          px2[ppair+pn]=sd->popsum[pn]*fx[0]*fx[0]*pinvn;
          py2[ppair+pn]=sd->popsum[pn]*fx[1]*fx[1]*pinvn;
          pz2[ppair+pn]=sd->popsum[pn]*fx[2]*fx[2]*pinvn;
          pxy[ppair+pn]=sd->popsum[pn]*fx[0]*fx[1]*pinvn;
          pxz[ppair+pn]=sd->popsum[pn]*fx[0]*fx[2]*pinvn;
          pyz[ppair+pn]=sd->popsum[pn]*fx[1]*fx[2]*pinvn;
          /*
          px2[pn]=sd->popsum[pn]*mcr->nodeid; // TEST
          */
        }
        // epop

	    fa+=3;

    }//while fa < nfa

    //fprintf(stderr,"Just before MPI sum: x2[0]=%7.3f\n",x2[0]);fflush(stderr);

//#ifdef GMX_MPI
    if (sd->M > 1)
    {
        gmx_sum_sim(sd->nres,x2,ms);
        gmx_sum_sim(sd->nres,y2,ms);
        gmx_sum_sim(sd->nres,z2,ms);
        gmx_sum_sim(sd->nres,xy,ms);
        gmx_sum_sim(sd->nres,xz,ms);
        gmx_sum_sim(sd->nres,yz,ms);

        // pop
        gmx_sum_sim(nrpop,px2,ms);
        gmx_sum_sim(nrpop,py2,ms);
        gmx_sum_sim(nrpop,pz2,ms);
        gmx_sum_sim(nrpop,pxy,ms);
        gmx_sum_sim(nrpop,pxz,ms);
        gmx_sum_sim(nrpop,pyz,ms);
        // epop


    }
//#endif
    //fprintf(stderr,"After MPI sum: x2[0]=%7.3f\n",x2[0]);fflush(stderr);

    fa=0;
    pair=0;
	while (fa < nfa){
      
	  type=forceatoms[fa];
      // population stuff back to usual variables
      ppair=pair*sd->popnum;
      x2[pair]=px2[ppair+sd->popid];
      y2[pair]=py2[ppair+sd->popid];
      z2[pair]=pz2[ppair+sd->popid];
      xy[pair]=pxy[ppair+sd->popid];
      xz[pair]=pxz[ppair+sd->popid];
      yz[pair]=pyz[ppair+sd->popid];
      /* TEST
      if (fa == 0){
        fprintf(stderr,"x2zero on node %d (popid %d) %7.3f\n",mcr->nodeid,sd->popid,x2[0]);
      }
      // TEST */



	    s2calc[pair]=(1.5/rijmin4[pair])*(x2[pair]*x2[pair] + y2[pair]*y2[pair] + z2[pair]*z2[pair] + 2*xy[pair]*xy[pair] + 2*xz[pair]*xz[pair] + 2*yz[pair]*yz[pair])-0.5;

        //if (((sd->s2step % sd->s2reout) == 0) && MASTER(cr))
        if ((sd->s2reout > 0) && (sd->s2step % sd->s2reout) == 0)
              fprintf (fplog,"S2RES: step %d : S2 for pair %d (atoms %d and %d) = %7.3f EXP: %7.3f\n", sd->s2step,pair,forceatoms[fa+1],forceatoms[fa+2],s2calc[pair],sd->s2exp[pair]);

	    fa += 3;
	    pair++;

    }

}//s2_calc


real ta_s2res(int nfa,const t_iatom forceatoms[],const t_iparams ip[],
               const rvec x[],rvec f[],rvec fshift[],
               const t_pbc *pbc,const t_graph *g,
               real lambda,real *dvdlambda,
               const t_mdatoms *md,t_fcdata *fcd,
               int *global_atom_index)
{

//    fprintf(stderr,"Maxwell's silver hammer\n");




    //return 0.0;
    //}


    t_s2resdata *sd;
    sd=&(fcd->s2res);
    sd->rho=0;
    real rho=sd->rho;
    real rhomin=sd->rhomin;
    real *S2=sd->s2exp; 
    real *S2err=sd->s2err;
    int type, ai,aj;
    real vtot=0.0; //fij=0.0;

    rvec rfij, fij;

    int ki=CENTRAL;
    rvec dx;
    ivec dt;

    real *x2=sd->x2;
    real *y2=sd->y2;
    real *z2=sd->z2;
    real *xy=sd->xy;
    real *xz=sd->xz;
    real *yz=sd->yz;

    rvec *ijv=sd->ijvector;
    real *rijmin4=sd->rijmin4;

    real *s2diff = sd->s2diff;


    // alpha is a global thing, not a per restraint measure
    //real alpha=500.00;
    real alpha=sd->alpha;
    //fprintf (stderr,"ALPHA=%7.3f\n",alpha);fflush(stderr);

    real *s2calc=sd->s2calc;

    int M=sd->M; //conforming the notation in Best & Vendruscolo
    //int M=sd->pop; //not populating for now
    int Ne=sd->nres;
    int pair, fa, m;

    fa=0;
    pair=0;
    rho=0.0;
    tensor invR;
/*
    fprintf(stderr,"xx %7.3f yy %7.3f zz %7.3f\n",sd->R[0][0],sd->R[0][1],sd->R[0][2]);fflush(stderr);
    fprintf(stderr,"xx %7.3f yy %7.3f zz %7.3f\n",sd->R[1][0],sd->R[1][1],sd->R[1][2]);fflush(stderr);
    fprintf(stderr,"xx %7.3f yy %7.3f zz %7.3f\n",sd->R[2][0],sd->R[2][1],sd->R[2][2]);fflush(stderr);
*/
    if (sd->do_fit){    
      // Matrix to rotate force components back 
      m_inv(sd->R,invR);

    }

    clear_rvec(rfij);
    clear_rvec(fij);

    while(fa < nfa){

    	if (s2calc[pair] < S2[pair]-S2err[pair]){
	      s2diff[pair]=s2calc[pair]-(S2[pair]-S2err[pair]);
	    } else if (s2calc[pair] > S2[pair]+S2err[pair]){
	      s2diff[pair]=s2calc[pair]-(S2[pair]+S2err[pair]);
	    } else{
	      s2diff[pair]=0.0;
	    }
        //fprintf(stderr,"pair %d s2 %7.3f s2calc %7.3f s2diff %7.3f\n",pair, S2[pair],s2calc[pair],s2diff[pair]);fflush(stderr);

        rho+=(s2diff[pair]*s2diff[pair]);


        fa+=3;
        pair++;
    }//while fa<nfa

    rho/=Ne;
    sd->rho=rho;

    if (rhomin < 0.0){
        sd->rhomin=rho;
	vtot=0.0;
    } else if (rho > rhomin){
	vtot=(alpha*M/2)*(rho-rhomin)*(rho-rhomin);


	fa=0;
	pair=0;
	while(fa < nfa){	  // Calculating force only whne there is violation for that restraint
	  //if ((s2calc[pair] < S2[pair]-S2err[pair]) || (s2calc[pair] > S2[pair]+S2err[pair])){
	    ai   = forceatoms[fa+1];
	    aj   = forceatoms[fa+2];

	    if (pbc)
		   ki=pbc_dx_aiuc(pbc,x[ai],x[aj],dx);
           //ki=pbc_dx(pbc,x[ai],x[aj],dx);
        else
            rvec_sub(x[ai],x[aj],dx);

	    if (g) {
		    ivec_sub(SHIFT_IVEC(g,ai),SHIFT_IVEC(g,aj),dt);
		    ki=IVEC2IS(dt);
	    }

	    rfij[0]=(12*alpha/Ne)*(rho-rhomin)*(s2diff[pair]/rijmin4[pair])*(x2[pair]*ijv[pair][0]+xy[pair]*ijv[pair][1]+xz[pair]*ijv[pair][2]);
            //Y
	    rfij[1]=(12*alpha/Ne)*(rho-rhomin)*(s2diff[pair]/rijmin4[pair])*(y2[pair]*ijv[pair][1]+yz[pair]*ijv[pair][2]+xy[pair]*ijv[pair][0]);
            //Z
	    rfij[2]=(12*alpha/Ne)*(rho-rhomin)*(s2diff[pair]/rijmin4[pair])*(z2[pair]*ijv[pair][2]+xz[pair]*ijv[pair][0]+yz[pair]*ijv[pair][1]);

            // If there was fitting, rotate force components back, otherwise simply copy them 
            if (sd->do_fit){
	      mvmul(invR,rfij,fij);
	    } else {
//#ifdef GMX_DOUBLE
//	      copy_dvec(rfij,fij);  
//#else
	      copy_rvec(rfij,fij);  
//#endif
	    }            

	    for(m=0; m<DIM; m++) {
	      f[ai][m]      += fij[m];
	      f[aj][m]      -= fij[m];
            //f[ai][m]      -= fij[m];
            //f[aj][m]      += fij[m];
	      fshift[ki][m] += fij[m];
	      fshift[CENTRAL][m] -= fij[m];
	    }

	    fa+=3;
	    pair++;

	}

    //fprintf(stderr,"FORCE! pair %d (atoms %d %d) F %7.3f %7.3f %7.3f (rho %7.3f, rhomin %7.3f, vtot %7.3f)\n",pair,ai,aj,fij[0],fij[1],fij[2],rho,rhomin,vtot);fflush(stderr);
        //sd->rhomin=sd->rho;


    }
    if (rho < rhomin)
	sd->rhomin=rho;
    
    return vtot;


}


