/*
 * 
 *                This source code is part of
 * 
 *                 G   R   O   M   A   C   S
 * 
 *          GROningen MAchine for Chemical Simulations
 * 
 *                        VERSION 3.2.0
 * Written by David van der Spoel, Erik Lindahl, Berk Hess, and others.
 * Copyright (c) 1991-2000, University of Groningen, The Netherlands.
 * Copyright (c) 2001-2004, The GROMACS development team,
 * check out http://www.gromacs.org for more information.

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * If you want to redistribute modifications, please consider that
 * scientific software is very special. Version control is crucial -
 * bugs must be traceable. We will be happy to consider code for
 * inclusion in the official distribution, but derived work must not
 * be called official GROMACS. Details are found in the README & COPYING
 * files - if they are missing, get the official version at www.gromacs.org.
 * 
 * To help us fund GROMACS development, we humbly ask that you cite
 * the papers on the package - you can find them in the top README file.
 * 
 * For more info, check our website at http://www.gromacs.org
 * 
 * And Hey:
 * GRoups of Organic Molecules in ACtion for Science
 */

#include "commrec.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef real rvec5[5];

/* Distance restraining stuff */
typedef struct {
  int  dr_weighting;  /* Weighting of pairs in one restraint              */
  gmx_bool dr_bMixed;     /* Use sqrt of the instantaneous times              *
		       * the time averaged violation                      */
  real dr_fc;	      /* Force constant for disres,                       *
		       * which is multiplied by a (possibly)              *
		       * different factor for each restraint              */
  real dr_tau;	      /* Time constant for disres		          */
  real ETerm;         /* multiplication factor for time averaging         */
  real ETerm1;        /* 1 - ETerm1                                       */
  real exp_min_t_tau; /* Factor for slowly switching on the force         */
  int  nres;          /* The number of distance restraints                */
  int  npair;         /* The number of distance restraint pairs           */
  real sumviol;       /* The sum of violations                            */
  real *rt;           /* The calculated instantaneous distance (npr)      */
  real *rm3tav;       /* The calculated time averaged distance (npr)      */
  real *Rtl_6;        /* The calculated instantaneous r^-6 (nr)           */
  real *Rt_6;         /* The calculated inst. ens. averaged r^-6 (nr)     */
  real *Rtav_6;       /* The calculated time and ens. averaged r^-6 (nr)  */
  int  nsystems;      /* The number of systems for ensemble averaging     */
  int  popsize;       /* Poplulation size   */
  gmx_bool pairwise;  /* Use pairwise averaging over the replicas  - ZG   */
  int  *leftmask;     /* mask for pairwise averaging */
  int  *rightmask;    /* mask for pairwise averaging */
  real *Rtleftmp;     /* Temporary array for over-replica summing with left neighbour (for MUMO) (nr)     */
  real *Rtrightmp;    /* Temporary array for over-replica summing with the right neighbour (for MUMO) (nr)     */
  MPI_Comm mpi_comm_ensemble; /* For ensemble averaging                   */
} t_disresdata;

/* Half-harmonic distance restraints with MUMO support */
typedef struct {
  int  hhdr_weighting;/* Weighting of pairs in one restraint              */
  real hhdr_tau;      /* Time constant for disres		          */
  gmx_bool hhdr_bMixed;     /* Use sqrt of the instantaneous times..      */
  gmx_bool pairwise;     /* Use pairwise averaging over the replicas      */
  real ETerm;         /* multiplication factor for time averaging         */
  real ETerm1;        /* 1 - ETerm1                                       */
  real exp_min_t_tau; /* Factor for slowly switching on the force         */
  int  nres;          /* The number of distance restraints                */
  int  npair;         /* The number of distance restraint pairs           */
  real sumviol;       /* The sum of violations                            */
  real hhdr_exp;      /* Exponent for ensemble averaging                  */
  real hhdr_fc;       /* Exponent for ensemble averaging                  */
  real *rt;           /* The calculated instantaneous distance (npr)      */
  real *rm3tav;       /* The calculated time averaged distance (npr)      */
  real *Rtl_6;        /* The calculated instantaneous r^-6 (nr)           */
  real *Rt_6;         /* The calculated inst. ens. averaged r^-6 (nr)     */
  real *Rt_left;      /* The calculated inst. averaged r^-6 with the left neighbour (for MUMO) (nr)     */
  real *Rt_right;     /* The calculated inst. averaged r^-6 with the right neighbour (for MUMO) (nr)     */
  real *Rtleftmp;     /* Temporary array for over-replica summing with left neighbour (for MUMO) (nr)     */
  real *Rtrightmp;    /* Temporary array for over-replica summing with the right neighbour (for MUMO) (nr)     */
  real *rnij;         /* The calculated simple average pair distance for each restraint   */
  real *Rtav_6;       /* The calculated time and ens. averaged r^-6 (nr)  */
  real *viol;         /* Individual violations  */
  real *violl;        /* Individual violations  */
  real *violr;        /* Individual violations  */
  real rho;           /* The progress variable rho */
  real rhomin;        /* The progress variable rho */
  int  nsystems;      /* The number of systems for ensemble averaging     */
  int  M;             /* The number of replicas for ensemble averaging    */
  int  pop;           /* population size for pairwise averaging */
  int  popid;         /* population id for pairwise averaging */
  int  leftn;         /* number of left neighbor for pairwise averaging */
  int  rightn;        /* number of right neighbor for pairwise averaging */
  int  popnum;        /* population number for pairwise averaging */
  int  *leftmask;     /* mask for pairwise averaging */
  int  *rightmask;    /* mask for pairwise averaging */
  MPI_Comm mpi_comm_ensemble; /* For ensemble averaging                   */
} t_hhdisresdata;


/* S2 restraining stuff (ZG) */
typedef struct {
  int nres;             /* number of S2 restraints */
  int npair;            /* number of atom pairs in restraints */
  real alpha;         /* alpha :-) */
  rvec *ijvector;     /* The vector between atoms in and j for whch the
                         order parameter is to be calculated              */

  int M;            /* Number of replicas (== nodes) */
  real *rijmin4;      /* Optimal bond length */
  real sumviol;       /* The sum of violations                            */
  real rho;           /* The progress variable rho */
  real rhomin;        /* The progress variable rho */
  real *s2calc;       /* The calculated S2*/
  real *s2exp;        /* Experimental S2*/
  real *s2err;        /* Error of experimental S2*/
  real *s2diff;       /* Difference between the restrained and calculated S2*/
  real *x2;
  real *y2;
  real *z2;
  real *xy;
  real *xz;
  real *yz;
  int  s2reout;       /* print frequency of S2 values to log file (=nsts2reout) */
  int  s2step;        /* ugly but useful: a separate step counter until I can master energy logging */
  int  pop;           /* population size */
  int  popid;         /* population id of node */
  int  popnum;        /* number of populations */
  short *popsum;      /* array of size popnum, holds 1 for popid, 0 for others */
  real *px2;          /* developmental stuff for population calc, later might completely replace x2, y2 etc... */
  real *py2;
  real *pz2;
  real *pxy;
  real *pxz;
  real *pyz;

  /* Fitting stuff from orires part */
  int  nref;          /* The number of atoms for the fit                    */
  real invmref;       /* The inverse total mass of the fit atoms            */
  real *mref;         /* The masses of the reference atoms                  */
  rvec *xref;         /* The reference coordinates for the fit (nref)       */
  rvec *xtmp;         /* Temporary array for fitting (nref)                 */
  tensor R;           /* Best-fit tensor */
  int do_fit;        /* my invention, hope will work */

  MPI_Comm mpi_comm_ensemble; /* For ensemble averaging */ // might also be needed in this version (see disres)

} t_s2resdata;
/* GZ */

/* Orientation restraining stuff */
typedef struct {
  real   fc;          /* Force constant for the restraints                  */
  real   edt;         /* Multiplication factor for time averaging           */
  real   edt_1;        /* 1 - edt                                            */
  real   exp_min_t_tau; /* Factor for slowly switching on the force         */
  int    nr;          /* The number of orientation restraints               */
  int    nex;         /* The number of experiments                          */
  int  nref;          /* The number of atoms for the fit                    */
  real *mref;         /* The masses of the reference atoms                  */
  rvec *xref;         /* The reference coordinates for the fit (nref)       */
  rvec *xtmp;         /* Temporary array for fitting (nref)                 */
  matrix R;           /* Rotation matrix to rotate to the reference coor.   */
  tensor *S;          /* Array of order tensors for each experiment (nexp)  */
  rvec5  *Dinsl;      /* The order matrix D for all restraints (nr x 5)     */
  rvec5  *Dins;       /* The ensemble averaged D (nr x 5)                   */
  rvec5  *Dtav;       /* The time and ensemble averaged D (nr x 5)          */
  real   *oinsl;      /* The calculated instantaneous orientations          */
  real   *oins;       /* The calculated emsemble averaged orientations      */
  real   *otav;       /* The calculated time and ensemble averaged orient.  */
  real   rmsdev;      /* The weighted (using kfac) RMS deviation            */
  rvec5  *tmp;        /* An array of temporary 5-vectors (nex);             */ 
  real   ***TMP;      /* An array of temporary 5x5 matrices (nex);          */
  real   *eig;        /* Eigenvalues/vectors, for output only (nex x 12)    */

  /* variables for diagonalization with diagonalize_orires_tensors()*/
  double **M;
  double *eig_diag;
  double **v;
} t_oriresdata;

/* 
 * Data struct used in the force calculation routines
 * for storing the tables for bonded interactions and
 * for storing information which is needed in following steps
 * (for instance for time averaging in distance retraints)
 * or for storing output, since force routines only return the potential.
 */
typedef struct {
  bondedtable_t *bondtab;
  bondedtable_t *angletab;
  bondedtable_t *dihtab;

  t_disresdata disres;
  t_hhdisresdata hhdisres; // ZG
  t_oriresdata orires;
  t_s2resdata  s2res; // ZG
  real         dihre_fc;
} t_fcdata;

#ifdef __cplusplus
}
#endif

